@@ -4,7 +4,7 @@
  * Copyright (c) 2003 Marius Aamodt Eriksen <marius@monkey.org>
  * All rights reserved.
  *
- * $Id: trickledu.h,v 1.6 2003/04/15 05:44:54 marius Exp $
+ * $Id: trickledu.h,v 1.7 2004/02/13 06:11:21 marius Exp $
  */
 
 #ifndef TRICKLE_TRICKLEDU_H
@@ -12,7 +12,8 @@
 
 void            trickled_configure(char *, int (*)(int, int, int),
                     ssize_t (*)(int, void *, size_t),
-                    ssize_t (*)(int, const void *, size_t), char *);
+                    ssize_t (*)(int, const void *, size_t),
+		    int (*)(int), char *);
 void            trickled_open(int *);
 int             trickled_update(short, size_t);
 int             trickled_delay(short, size_t *);
@@ -22,5 +23,6 @@ int             trickled_recvmsg(struct msg *);
 void            trickled_ctl_open(int *);
 int             trickled_getinfo(uint32_t *, uint32_t *,
                     uint32_t *, uint32_t *);
+void            trickled_close(int *);
 
 #endif /* TRICKLE_TRICKLEDU_H */
